from abc import ABCMeta  # type: ignore
from abc import abstractmethod
import os
import re
import typing as t

import pluggy  # type: ignore

import abnosql.exceptions as ex
from abnosql import plugin

hookimpl = pluggy.HookimplMarker('abnosql.table')
hookspec = pluggy.HookspecMarker('abnosql.table')


class TableSpecs(plugin.PluginSpec):

    @hookspec(firstresult=True)
    def set_config(self, table: str) -> t.Dict:  # type: ignore[empty-body] # noqa E501
        pass

    @hookspec(firstresult=True)
    def get_item_post(self, table: str, item: t.Dict) -> t.Dict:  # type: ignore[empty-body] # noqa E501
        pass

    @hookspec
    def put_item_post(self, table: str, item: t.Dict) -> None:  # type: ignore[empty-body] # noqa E501
        pass

    @hookspec
    def put_items_post(self, table: str, items: t.List[t.Dict]) -> None:  # type: ignore[empty-body] # noqa E501
        pass

    @hookspec
    def delete_item_post(self, table: str, key: t.Dict) -> None:  # type: ignore[empty-body] # noqa E501
        pass


class TableBase(metaclass=ABCMeta):
    @abstractmethod
    def __init__(
        self, pm: plugin.PM, name: str, config: t.Optional[dict] = None
    ) -> None:
        pass

    @abstractmethod
    def get_item(self, **kwargs) -> t.Dict:
        pass

    @abstractmethod
    def put_item(self, item: t.Dict):
        pass

    @abstractmethod
    def put_items(self, items: t.Iterable[t.Dict]):
        pass

    @abstractmethod
    def delete_item(self, **kwargs):
        pass

    @abstractmethod
    def query(
        self,
        key: t.Dict[str, t.Any],
        filters: t.Optional[t.Dict[str, t.Any]] = None,
        limit: t.Optional[int] = None,
        next: t.Optional[str] = None
    ) -> t.Dict[str, t.Any]:
        pass

    @abstractmethod
    def query_sql(
        self,
        statement: str,
        parameters: t.Optional[t.Dict[str, t.Any]] = None,
        limit: t.Optional[int] = None,
        next: t.Optional[str] = None
    ) -> t.Dict[str, t.Any]:
        pass


def get_sql_params(
    statement: str,
    parameters: t.Dict[str, t.Any],
    param_val: t.Callable,
    replace: t.Optional[str] = None
) -> t.Tuple[str, t.List]:
    # convert @variable to dynamodb ? placeholders
    validate_statement(statement)
    vars = list(re.findall(r'\@[a-zA-Z0-9_.-]+', statement))
    params = []
    _missing = {}
    for var in vars:
        if var not in parameters:
            _missing[var] = True
        else:
            val = parameters[var]
            params.append(param_val(var, val))
    for var in parameters.keys():
        if var not in vars:
            _missing[var] = True
    missing = sorted(_missing.keys())
    if len(missing):
        raise ex.ValidationException(
            'missing parameters: ' + ', '.join(missing)
        )
    if isinstance(replace, str):
        for var in parameters.keys():
            statement = statement.replace(var, replace)
    return (statement, params)


def quote_str(str):
    return "'" + str.translate(
        str.maketrans({
            "'": "\\'"
        })
    ) + "'"


def validate_query_attrs(key: t.Dict, filters: t.Dict):
    _name_pat = re.compile(r'^[a-zA-Z09_-]+$')

    def _validate_key_names(obj):
        return [_ for _ in obj.keys() if not _name_pat.match(_)]

    invalid = sorted(set(
        _validate_key_names(key) + _validate_key_names(filters)
    ))
    if len(invalid):
        raise ex.ValidationException(
            'invalid key or filter keys: ' + ', '.join(invalid)
        )


def validate_statement(statement: str):
    # sqlglot can do this (and sqlparse), but lets keep it simple
    tokens = [_.strip() for _ in statement.split(' ') if _.strip() != '']
    if len(tokens) == 0 or tokens[0].upper() != 'SELECT':
        raise ex.ValidationException('statement must start with SELECT')


def table(
    name: str,
    config: t.Optional[dict] = None,
    database: t.Optional[str] = None
) -> TableBase:
    if database is None:
        database = os.environ.get('ABNOSQL_DB')
    pm = plugin.get_pm('table')
    module = pm.get_plugin(database)
    if module is None:
        raise ex.PluginException(f'table.{database} plugin not found')
    return module.Table(pm, name, config)



import functools
import json
from unittest.mock import patch

import boto3  # type: ignore
import botocore  # type: ignore
from dynamodb_json import json_util  # type: ignore

from abnosql.plugins.table.dynamodb import deserialize
from abnosql.plugins.table.memory import get_table_name
from abnosql.plugins.table.memory import query_items

ORIG_MAKE_API_CALL = botocore.client.BaseClient._make_api_call


def mock_dynamodbx(f):

    # won't need this when moto supports ExecuteStatement
    def execute_statement(kwargs):
        table_name = get_table_name(kwargs['Statement'])
        table = boto3.resource('dynamodb').Table(table_name)
        all_items = deserialize(table.scan()['Items'])
        items = []
        _items = query_items(
            kwargs['Statement'],
            all_items,
            kwargs.get('Parameters')
        )
        for item in _items:
            items.append(json.loads(json_util.dumps(item)))
        return {
            'Items': items
        }

    FUNC_MAP = {
        'ExecuteStatement': execute_statement
    }

    def _mock(self, operation_name, kwargs):
        response = None
        mock_func = FUNC_MAP.get(operation_name)
        if callable(mock_func):
            response = mock_func(kwargs)
            if response is None:
                raise Exception('mock_dynamodbx operation %s not supported' % (
                    operation_name
                ))
        else:
            response = ORIG_MAKE_API_CALL(self, operation_name, kwargs)
        return response

    @functools.wraps(f)
    def decorated(*args, **kwargs):
        _make_api_call = 'botocore.client.BaseClient._make_api_call'
        with patch(_make_api_call, _mock):
            return f(*args, **kwargs)
    return decorated



#!/usr/bin/env python
import os
from setuptools import find_packages  # type: ignore
from setuptools import setup


PACKAGE_DIR = os.path.abspath(os.path.dirname(__file__))


def read(*args):
    return open(os.path.join(PACKAGE_DIR, *args)).read()


def get_version():
    line = read('abnosql', 'version.py').splitlines()[0].strip()
    parts = line.split(' ')
    assert len(parts) == 3
    assert parts[0] == '__version__'
    assert parts[1] == '='
    return parts[2].strip('\'"')


__version__ = get_version()


base_deps = [
    'click',
    'pluggy',
    'sqlglot',
    'tabulate'
]
dynamodb_deps = [
    'boto3',
    'dynamodb_json'
]
cosmos_deps = [
    'azure-cosmos'
]
all_deps = base_deps + dynamodb_deps + cosmos_deps
tests_require = all_deps + [
    'coverage',
    'moto[dynamodb]',
    'mypy',
    'pytest',
    'pytest-cov',
    'responses'
]
dev_require = tests_require + [
    'pre-commit'
]

setup(
    name='abnosql',
    version=__version__,
    description='NoSQL Abstraction Library',
    long_description=read('README.md'),
    long_description_content_type="text/markdown",
    author='Roger Foskett',
    author_email='r_foskett@hotmail.com',
    maintainer='Roger Foskett',
    maintainer_email='r_foskett@hotmail.com',

    url='https://github.com/rog555/abnosql',
    download_url='http://pypi.python.org/pypi/abnosql',
    keywords='nosql, azure cosmos, aws dynamodb',

    license='MIT',
    platforms='any',
    packages=find_packages(exclude=['tests']),

    tests_require=tests_require,
    extras_require={
        'dev': dev_require,
        'test': tests_require,
        'dynamodb': dynamodb_deps,
        'cosmos': cosmos_deps
    },
    python_requires='>=3.8,<4.0',
    test_suite='tests',

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: System :: Distributed Computing',
        'Topic :: Database :: Front-Ends',
        'Typing :: Typed'
    ],
    entry_points={
        'console_scripts': [
            'abnosql = abnosql:cli',
        ]
    }
)
